@@ -1,4 +1,4 @@
-use std::path::Path;
+use std::{path::Path, sync::Arc};
 
 use indexmap::IndexMap;
 use log::trace;
@@ -62,8 +62,8 @@ pub fn parse_simple_column_path(
                 output.push(Member::Bare(trimmed.clone().spanned(part_span)));
             }
             current_part.clear();
-            // Note: I believe this is safe because of the delimiter we're using, but if we get fancy with
-            // unicode we'll need to change this
+            // Note: I believe this is safe because of the delimiter we're using,
+            // but if we get fancy with Unicode we'll need to change this.
             start_index = idx + '.'.len_utf8();
             continue;
         }
@@ -143,8 +143,8 @@ pub fn parse_full_column_path(
                 );
             }
             current_part.clear();
-            // Note: I believe this is safe because of the delimiter we're using, but if we get fancy with
-            // unicode we'll need to change this
+            // Note: I believe this is safe because of the delimiter we're using,
+            // but if we get fancy with Unicode we'll need to change this.
             start_index = idx + '.'.len_utf8();
             continue;
         }
@@ -227,6 +227,15 @@ fn parse_range(
         );
     };
 
+    if lite_arg.item[0..dotdot_pos].is_empty()
+        && lite_arg.item[(dotdot_pos + operator_str.len())..].is_empty()
+    {
+        return (
+            garbage(lite_arg.span),
+            Some(ParseError::mismatch("range", lite_arg.clone())),
+        );
+    }
+
     let numbers: Vec<_> = lite_arg.item.split(operator_str).collect();
 
     if numbers.len() != 2 {
@@ -325,51 +334,74 @@ fn parse_operator(lite_arg: &Spanned<String>) -> (SpannedExpression, Option<Pars
 
 /// Parse a unit type, eg '10kb'
 fn parse_unit(lite_arg: &Spanned<String>) -> (SpannedExpression, Option<ParseError>) {
+    fn parse_decimal_str_to_number(decimal: &str) -> Option<i64> {
+        let string_to_parse = format!("0.{}", decimal);
+        if let Ok(x) = string_to_parse.parse::<f64>() {
+            return Some((1_f64 / x) as i64);
+        }
+        None
+    }
     let unit_groups = [
-        (Unit::Byte, vec!["b", "B"]),
-        (Unit::Kilobyte, vec!["kb", "KB", "Kb", "kB"]),
-        (Unit::Megabyte, vec!["mb", "MB", "Mb", "mB"]),
-        (Unit::Gigabyte, vec!["gb", "GB", "Gb", "gB"]),
-        (Unit::Terabyte, vec!["tb", "TB", "Tb", "tB"]),
-        (Unit::Petabyte, vec!["pb", "PB", "Pb", "pB"]),
-        (Unit::Kibibyte, vec!["KiB", "kib", "kiB", "Kib"]),
-        (Unit::Mebibyte, vec!["MiB", "mib", "miB", "Mib"]),
-        (Unit::Gibibyte, vec!["GiB", "gib", "giB", "Gib"]),
-        (Unit::Nanosecond, vec!["ns"]),
-        (Unit::Microsecond, vec!["us"]),
-        (Unit::Millisecond, vec!["ms"]),
-        (Unit::Second, vec!["sec"]),
-        (Unit::Minute, vec!["min"]),
-        (Unit::Hour, vec!["hr"]),
-        (Unit::Day, vec!["day"]),
-        (Unit::Week, vec!["wk"]),
-        (Unit::Month, vec!["mon"]),
-        (Unit::Year, vec!["yr"]),
+        (Unit::Kilobyte, "KB", Some((Unit::Byte, 1000))),
+        (Unit::Megabyte, "MB", Some((Unit::Kilobyte, 1000))),
+        (Unit::Gigabyte, "GB", Some((Unit::Megabyte, 1000))),
+        (Unit::Terabyte, "TB", Some((Unit::Gigabyte, 1000))),
+        (Unit::Petabyte, "PB", Some((Unit::Terabyte, 1000))),
+        (Unit::Kibibyte, "KIB", Some((Unit::Byte, 1024))),
+        (Unit::Mebibyte, "MIB", Some((Unit::Kibibyte, 1024))),
+        (Unit::Gibibyte, "GIB", Some((Unit::Mebibyte, 1024))),
+        (Unit::Tebibyte, "TIB", Some((Unit::Gibibyte, 1024))),
+        (Unit::Pebibyte, "PIB", Some((Unit::Tebibyte, 1024))),
+        (Unit::Byte, "B", None),
+        (Unit::Nanosecond, "NS", None),
+        (Unit::Microsecond, "US", Some((Unit::Nanosecond, 1000))),
+        (Unit::Millisecond, "MS", Some((Unit::Microsecond, 1000))),
+        (Unit::Second, "SEC", Some((Unit::Millisecond, 1000))),
+        (Unit::Minute, "MIN", Some((Unit::Second, 60))),
+        (Unit::Hour, "HR", Some((Unit::Minute, 60))),
+        (Unit::Day, "DAY", Some((Unit::Minute, 1440))),
+        (Unit::Week, "WK", Some((Unit::Day, 7))),
     ];
+    if let Some(unit) = unit_groups
+        .iter()
+        .find(|&x| lite_arg.to_uppercase().ends_with(x.1))
+    {
+        let mut lhs = lite_arg.item.clone();
+        for _ in 0..unit.1.len() {
+            lhs.pop();
+        }
 
-    for unit_group in unit_groups.iter() {
-        for unit in unit_group.1.iter() {
-            if !lite_arg.item.ends_with(unit) {
-                continue;
-            }
-            let mut lhs = lite_arg.item.clone();
-
-            for _ in 0..unit.len() {
-                lhs.pop();
-            }
-
-            // these units are allowed to be signed
-            if let Ok(x) = lhs.parse::<i64>() {
-                let lhs_span = Span::new(lite_arg.span.start(), lite_arg.span.start() + lhs.len());
-                let unit_span = Span::new(lite_arg.span.start() + lhs.len(), lite_arg.span.end());
-                return (
-                    SpannedExpression::new(
-                        Expression::unit(x.spanned(lhs_span), unit_group.0.spanned(unit_span)),
-                        lite_arg.span,
+        let input: Vec<&str> = lhs.split('.').collect();
+        let (value, unit_to_use) = match &input[..] {
+            [number_str] => (number_str.parse::<i64>().ok(), unit.0),
+            [number_str, decimal_part_str] => match unit.2 {
+                Some(unit_to_convert_to) => match (
+                    number_str.parse::<i64>(),
+                    parse_decimal_str_to_number(decimal_part_str),
+                ) {
+                    (Ok(number), Some(decimal_part)) => (
+                        Some(
+                            (number * unit_to_convert_to.1) + (unit_to_convert_to.1 / decimal_part),
+                        ),
+                        unit_to_convert_to.0,
                     ),
-                    None,
-                );
-            }
+                    _ => (None, unit.0),
+                },
+                None => (None, unit.0),
+            },
+            _ => (None, unit.0),
+        };
+
+        if let Some(x) = value {
+            let lhs_span = Span::new(lite_arg.span.start(), lite_arg.span.start() + lhs.len());
+            let unit_span = Span::new(lite_arg.span.start() + lhs.len(), lite_arg.span.end());
+            return (
+                SpannedExpression::new(
+                    Expression::unit(x.spanned(lhs_span), unit_to_use.spanned(unit_span)),
+                    lite_arg.span,
+                ),
+                None,
+            );
         }
     }
 
@@ -402,9 +434,13 @@ fn parse_invocation(
     };
 
     scope.enter_scope();
-    let (classified_block, err) = classify_block(&lite_block, scope);
+    let (mut classified_block, err) = classify_block(&lite_block, scope);
     scope.exit_scope();
 
+    if let Some(x) = std::sync::Arc::<nu_protocol::hir::Block>::get_mut(&mut classified_block) {
+        x.set_redirect(ExternalRedirection::Stdout);
+    }
+
     (
         SpannedExpression::new(Expression::Invocation(classified_block), lite_arg.span),
         err,
@@ -450,13 +486,11 @@ fn parse_dollar_expr(
         //Return invocation
         trace!("Parsing invocation expression");
         parse_invocation(lite_arg, scope)
-    } else if lite_arg.item.contains("..") {
-        parse_range(lite_arg, scope)
-    } else if lite_arg.item.contains('.') {
-        trace!("Parsing path expression");
-        parse_full_column_path(lite_arg, scope)
+    } else if let (expr, None) = parse_range(lite_arg, scope) {
+        (expr, None)
+    } else if let (expr, None) = parse_full_column_path(lite_arg, scope) {
+        (expr, None)
     } else {
-        trace!("Parsing variable expression");
         parse_variable(lite_arg, scope)
     }
 }
@@ -613,12 +647,12 @@ fn parse_interpolated_string(
     let group = Group::new(pipelines, lite_arg.span);
 
     let call = SpannedExpression {
-        expr: Expression::Invocation(Block::new(
+        expr: Expression::Invocation(Arc::new(Block::new(
             Signature::new("<invocation>"),
             vec![group],
             IndexMap::new(),
             lite_arg.span,
-        )),
+        ))),
         span: lite_arg.span,
     };
 
@@ -1261,7 +1295,7 @@ fn parse_positional_argument(
                 span,
             );
 
-            let arg = SpannedExpression::new(Expression::Block(block), span);
+            let arg = SpannedExpression::new(Expression::Block(Arc::new(block)), span);
 
             idx = new_idx - 1;
             if error.is_none() {
@@ -1304,7 +1338,7 @@ fn parse_positional_argument(
                         span,
                     );
 
-                    let arg = SpannedExpression::new(Expression::Block(block), span);
+                    let arg = SpannedExpression::new(Expression::Block(Arc::new(block)), span);
 
                     idx = new_idx - 1;
                     if error.is_none() {
@@ -1923,7 +1957,7 @@ fn parse_alias(call: &LiteCommand, scope: &dyn ParserScope) -> Option<ParseError
 pub fn classify_block(
     lite_block: &LiteBlock,
     scope: &dyn ParserScope,
-) -> (Block, Option<ParseError>) {
+) -> (Arc<Block>, Option<ParseError>) {
     let mut output = Block::basic();
     let mut error = None;
 
@@ -1987,7 +2021,7 @@ pub fn classify_block(
                         span: Span::new(vars.0.span.start(), vars.1.span.end()),
                     },
                     SpannedExpression {
-                        expr: Expression::Block(block),
+                        expr: Expression::Block(Arc::new(block)),
                         span,
                     },
                 ]);
@@ -2021,22 +2055,23 @@ pub fn classify_block(
             output.definitions.insert(name, definition.clone());
         }
     }
+    output.infer_params();
 
-    (output, error)
+    (Arc::new(output), error)
 }
 
 pub fn parse(
     input: &str,
     span_offset: usize,
     scope: &dyn ParserScope,
-) -> (Block, Option<ParseError>) {
+) -> (Arc<Block>, Option<ParseError>) {
     let (output, error) = lex(input, span_offset);
     if error.is_some() {
-        return (Block::basic(), error);
+        return (Arc::new(Block::basic()), error);
     }
     let (lite_block, error) = parse_block(output);
     if error.is_some() {
-        return (Block::basic(), error);
+        return (Arc::new(Block::basic()), error);
     }
 
     classify_block(&lite_block, scope)
@@ -2156,6 +2191,26 @@ fn unit_parse_byte_units() {
             value: 123,
             unit: Unit::Gibibyte,
         },
+        TestCase {
+            string: String::from("10tib"),
+            value: 10,
+            unit: Unit::Tebibyte,
+        },
+        TestCase {
+            string: String::from("123TiB"),
+            value: 123,
+            unit: Unit::Tebibyte,
+        },
+        TestCase {
+            string: String::from("10pib"),
+            value: 10,
+            unit: Unit::Pebibyte,
+        },
+        TestCase {
+            string: String::from("123PiB"),
+            value: 123,
+            unit: Unit::Pebibyte,
+        },
     ];
 
     for case in cases.iter() {
@@ -2179,3 +2234,91 @@ fn unit_parse_byte_units() {
         );
     }
 }
+
+#[test]
+fn unit_parse_byte_units_decimal() {
+    struct TestCase {
+        string: String,
+        value: i64,
+        value_str: String,
+        unit: Unit,
+    }
+
+    let cases = [
+        TestCase {
+            string: String::from("0.25KB"),
+            value: 250,
+            value_str: String::from("0.25"),
+            unit: Unit::Byte,
+        },
+        TestCase {
+            string: String::from("2.5Mb"),
+            value: 2500,
+            value_str: String::from("2.5"),
+            unit: Unit::Kilobyte,
+        },
+        TestCase {
+            string: String::from("0.5Gb"),
+            value: 500,
+            value_str: String::from("0.5"),
+            unit: Unit::Megabyte,
+        },
+        TestCase {
+            string: String::from("811.5Gb"),
+            value: 811500,
+            value_str: String::from("811.5"),
+            unit: Unit::Megabyte,
+        },
+        TestCase {
+            string: String::from("11.5Tb"),
+            value: 11500,
+            value_str: String::from("11.5"),
+            unit: Unit::Gigabyte,
+        },
+        TestCase {
+            string: String::from("12.5Pb"),
+            value: 12500,
+            value_str: String::from("12.5"),
+            unit: Unit::Terabyte,
+        },
+        TestCase {
+            string: String::from("10.5kib"),
+            value: 10752,
+            value_str: String::from("10.5"),
+            unit: Unit::Byte,
+        },
+        TestCase {
+            string: String::from("0.5mib"),
+            value: 512,
+            value_str: String::from("0.5"),
+            unit: Unit::Kibibyte,
+        },
+        TestCase {
+            string: String::from("3.25gib"),
+            value: 3328,
+            value_str: String::from("3.25"),
+            unit: Unit::Mebibyte,
+        },
+    ];
+
+    for case in cases.iter() {
+        let input_len = case.string.len();
+        let value_len = case.value_str.to_string().len();
+        let input = case.string.clone().spanned(Span::new(0, input_len));
+        let result = parse_unit(&input);
+        assert_eq!(result.1, None);
+        assert_eq!(
+            result.0.expr,
+            Expression::unit(
+                Spanned {
+                    span: Span::new(0, value_len),
+                    item: case.value
+                },
+                Spanned {
+                    span: Span::new(value_len, input_len),
+                    item: case.unit
+                }
+            )
+        );
+    }
+}
